"""
This file contains tests for fields.py
"""

import datetime
import json
import random
import re
import uuid
from datetime import timezone, date
from io import BytesIO
from unittest import TestCase

from django.core.exceptions import ValidationError
from django.core.validators import BaseValidator
from django.test import override_settings
from django.utils.timezone import now

from django_rest_form_fields.compatibility import to_timestamp
from django_rest_form_fields.fields import RestBooleanField, LowerCaseEmailField, TimestampField, DateUnitField, \
    ColorField, IdArrayField, IdSetField, TruncatedCharField, JsonField, ArrayField, UrlField, RestCharField, \
    RestChoiceField, RestIntegerField, RegexField, UUIDField, DateTimeField, MonthField, FileField, RestFloatField, \
    DateField, IdField


class TestErrorValidator(BaseValidator):
    """
    I use this validator to raise error in run_validators for test purposes
    """
    compare = lambda self, a, b: True  # noqa: E731
    clean = lambda self, x: 0  # noqa: E731
    message = "This validator always raises error, if run_validators is run"
    code = 'always_error'


class LowerCaseEmailFieldTest(TestCase):
    def test_to_lower(self):
        f = LowerCaseEmailField()
        self.assertEqual(f.clean('TeSt@mail.ru'), 'test@mail.ru')

    def test_required(self):
        f = LowerCaseEmailField(required=False)
        self.assertEqual(None, f.clean(None))

        f = LowerCaseEmailField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        f = LowerCaseEmailField(required=False, initial='test@mail.ru')
        self.assertEqual('test@mail.ru', f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = LowerCaseEmailField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class RestBooleanFieldTest(TestCase):
    def test_true(self):
        f = RestBooleanField()
        self.assertEqual(True, f.clean('true'))

    def test_1(self):
        f = RestBooleanField()
        self.assertEqual(True, f.clean('1'))

    def test_false(self):
        f = RestBooleanField()
        self.assertEqual(False, f.clean('false'))

    def test_0(self):
        f = RestBooleanField()
        self.assertEqual(False, f.clean('0'))

    def test_empty_string(self):
        f = RestBooleanField()
        self.assertEqual(False, f.clean(''))

    def test_not_empty_string(self):
        f = RestBooleanField()
        self.assertEqual(True, f.clean('some_text'))

    def test_None(self):
        f = RestBooleanField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_not_required_true(self):
        f = RestBooleanField(required=False)
        self.assertEqual(True, f.clean('true'))

    def test_not_required_1(self):
        f = RestBooleanField(required=False)
        self.assertEqual(True, f.clean('1'))

    def test_not_required_false(self):
        f = RestBooleanField(required=False)
        self.assertEqual(False, f.clean('false'))

    def test_not_required_0(self):
        f = RestBooleanField(required=False)
        self.assertEqual(False, f.clean('0'))

    def test_not_required_empty_string(self):
        f = RestBooleanField(required=False)
        self.assertEqual(False, f.clean(''))

    def test_not_required_not_empty_string(self):
        f = RestBooleanField(required=False)
        self.assertEqual(True, f.clean('some_text'))

    def test_not_required_None(self):
        f = RestBooleanField(required=False)
        self.assertEqual(None, f.clean(None))

    def test_initial_true(self):
        f = RestBooleanField(required=False, initial=True)
        self.assertEqual(True, f.clean(None))

    def test_initial_false(self):
        f = RestBooleanField(required=False, initial=False)
        self.assertEqual(False, f.clean(None))

    def test_required(self):
        f = RestBooleanField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = RestBooleanField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class RestIntegerFieldTest(TestCase):
    def test_correct(self):
        f = RestIntegerField()
        self.assertEqual(1, f.clean(1))
        f = RestIntegerField()
        self.assertEqual(123, f.clean(123))
        f = RestIntegerField()
        self.assertEqual(0, f.clean(0))
        f = RestIntegerField()
        self.assertEqual(-10, f.clean(-10))
        f = RestIntegerField()
        self.assertEqual(1, f.clean('1'))
        f = RestIntegerField()
        self.assertEqual(123, f.clean('123'))
        f = RestIntegerField()
        self.assertEqual(0, f.clean('0'))
        f = RestIntegerField()
        self.assertEqual(-10, f.clean('-10'))

    def test_required(self):
        f = RestIntegerField(required=True)
        with self.assertRaises(ValidationError):
            f.clean(None)

        f = RestIntegerField(required=False)
        self.assertEqual(None, f.clean(None))

    def test_initial(self):
        f = RestIntegerField(required=True, initial=123)
        with self.assertRaises(ValidationError):
            f.clean(None)

        f = RestIntegerField(required=False, initial=123)
        self.assertEqual(123, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = RestIntegerField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class IdFieldTest(TestCase):
    def test_string(self):
        f = IdField()
        self.assertEqual(1, f.clean('1'))

    def test_integer(self):
        f = IdField()
        self.assertEqual(2, f.clean(2))

    def test_zero_invalid(self):
        f = IdField()

        with self.assertRaises(ValidationError):
            f.clean("0")

    def test_negative(self):
        f = IdField()

        with self.assertRaises(ValidationError):
            f.clean("-1")

    def test_invalid_format(self):
        f = IdField()

        with self.assertRaises(ValidationError):
            f.clean("1.123")

    def test_with_zero(self):
        f = IdField(with_zero=True)
        self.assertEqual(0, f.clean('0'))

    def test_settings_max_value_not_set(self):
        f = IdField()
        self.assertEqual(2 ** 64 + 1, f.clean(2 ** 64 + 1))

    @override_settings(ID_FIELD_MAX_VALUE=2**64)
    def test_settings_max_value(self):
        f = IdField()
        with self.subTest("Max value"):
            self.assertEqual(2**64, f.clean(2**64))

        with self.subTest("Too big value"):
            with self.assertRaises(ValidationError):
                f.clean(2 ** 64 + 1)

    def test_max_value(self):
        f = IdField(max_value=100)
        with self.subTest("Max value"):
            self.assertEqual(100, f.clean(100))

        with self.subTest("Too big value"):
            with self.assertRaises(ValidationError):
                f.clean(101)

    def test_min_value(self):
        f = IdField(min_value=10)
        with self.subTest("Min value"):
            self.assertEqual(10, f.clean(10))

        with self.subTest("Too small value"):
            with self.assertRaises(ValidationError):
                f.clean(9)

    def test_required(self):
        with self.subTest("required=False"):
            f = IdField(required=False)
            self.assertIsNone(f.clean(None))

        with self.subTest("required=True"):
            f = IdField()

            with self.assertRaises(ValidationError):
                f.clean(None)

    def test_initial(self):
        f = IdArrayField(required=False, initial=1)
        self.assertEqual(1, f.clean(None))


class RestFloatFieldTest(TestCase):
    def test_correct(self):
        f = RestFloatField()
        self.assertEqual(1, f.clean(1))
        f = RestFloatField()
        self.assertEqual(123.456, f.clean(123.456))
        f = RestFloatField()
        self.assertEqual(0, f.clean(0))
        f = RestFloatField()
        self.assertEqual(-10.123, f.clean(-10.123))
        f = RestFloatField()
        self.assertEqual(1, f.clean('1'))
        f = RestFloatField()
        self.assertEqual(123, f.clean('123.0'))
        f = RestFloatField()
        self.assertEqual(0, f.clean('0'))
        f = RestFloatField()
        self.assertEqual(-10, f.clean('-10'))

    def test_required(self):
        f = RestFloatField(required=True)
        with self.assertRaises(ValidationError):
            f.clean(None)

        f = RestFloatField(required=False)
        self.assertEqual(None, f.clean(None))

    def test_initial(self):
        f = RestFloatField(required=True, initial=123.0)
        with self.assertRaises(ValidationError):
            f.clean(None)

        f = RestFloatField(required=False, initial=123.0)
        self.assertEqual(123, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = RestFloatField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class RestCharFieldTest(TestCase):
    def test_required(self):
        f = RestCharField(required=False)
        self.assertEqual(None, f.clean(None))

        f = RestCharField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = "test_string"
        f = RestCharField(required=False, initial=test_data)
        self.assertEqual(test_data, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = RestCharField(required=False, initial='', min_length=1)
        with self.assertRaises(ValidationError):
            f.clean('')

        f2 = RestCharField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f2.clean('')


class RegexFieldValidator(TestCase):
    SRE_MATCH_TYPE = type(re.match("", ""))

    def test_required(self):
        f = RegexField(required=False, regex=r'.*')
        self.assertIsNone(f.clean(None))

        f = RegexField(regex=r'.*')
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = "test_string"
        f = RegexField(required=False, initial=test_data, regex=r'.*')
        self.assertEqual(test_data, f.clean(None))

    def test_regex_valid(self):
        test_data = "test_string"
        f = RegexField(regex=r'^test.*$')
        self.assertEqual(test_data, f.clean(test_data))

    def test_regex_compiled(self):
        test_data = "test_string"
        f = RegexField(regex=re.compile('^test.*$'))
        self.assertEqual(test_data, f.clean(test_data))

    def test_regex_invalid(self):
        test_data = "test_string"
        f = RegexField(regex=r'^test1.*$')
        with self.assertRaises(ValidationError):
            f.clean(test_data)

    def test_init_assertions(self):
        with self.assertRaises(AssertionError):
            RegexField(regex=[])

        with self.assertRaises(AssertionError):
            RegexField(regex=r'', flags=None)

        with self.assertRaises(AssertionError):
            RegexField(regex=r'', flags=[])

    def test_flags(self):
        f = RegexField(regex=r'^test.*$')
        self.assertEqual("test_string", f.clean("test_string"))
        with self.assertRaises(ValidationError):
            f.clean("tESt_string")

        f = RegexField(regex=r'^test.*$', flags=re.I)
        self.assertEqual("test_string", f.clean("test_string"))
        self.assertEqual("tESt_string", f.clean("tESt_string"))

    def test_match(self):
        test_data = "test_string"
        f = RegexField(regex=r'^test(.*)$')
        f.clean(test_data)
        self.assertTrue(isinstance(f.match, self.SRE_MATCH_TYPE))
        self.assertEqual(f.match.group(0), test_data)
        self.assertEqual(f.match.group(1), "_string")

        f = RegexField(regex=r'^test(.*)$', required=False)
        f.clean(None)
        self.assertIsNone(f.match)

    def test_no_regex(self):
        # Acts as a simple CharField
        test_data = "test_string"
        f = RegexField()
        self.assertEqual(test_data, f.clean(test_data))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = RegexField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class RestChoiceFieldTest(TestCase):
    def test_choices_array(self):
        f = RestChoiceField(choices=["a", "b", "c"])
        self.assertEqual("a", f.clean("a"))

    def test_choices_tuples(self):
        f = RestChoiceField(choices=[("a", "a"), ("b", "b"), ("c", "c")])
        self.assertEqual("c", f.clean("c"))

    def test_choices_mixed(self):
        f = RestChoiceField(choices=[("a", "a"), "b", ("c", "c")])
        self.assertEqual("b", f.clean("b"))

    def test_invalid(self):
        f = RestChoiceField(choices=[("a", "a"), "b", ("c", "c")])
        with self.assertRaises(ValidationError):
            f.clean("d")

    def test_empty_string(self):
        f = RestChoiceField(required=False)
        self.assertEqual('', f.clean(''))

        f = RestChoiceField(required=False, choices=['a', 'b', 'c'])
        with self.assertRaises(ValidationError):
            f.clean('')

    def test_required(self):
        f = RestChoiceField(required=False)
        self.assertEqual(None, f.clean(None))

        f = RestChoiceField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = "test_string"
        f = RestChoiceField(required=False, initial=test_data)
        self.assertEqual(test_data, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = RestChoiceField(validators=[TestErrorValidator(0)], choices=['abc', ''])
        with self.assertRaises(ValidationError):
            f.clean('')


class TimestampFieldTest(TestCase):
    def test_now(self):
        now_dt = now().replace(microsecond=0)
        now_ts = to_timestamp(now_dt)
        f = TimestampField()
        res = f.clean(now_ts)
        self.assertEqual(now_dt, res)

    def test_timezones(self):
        dt = datetime.datetime(2017, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
        ts = to_timestamp(dt)
        f = TimestampField()
        res = f.clean(ts)
        self.assertEqual(dt, res)

        dt = datetime.datetime(2017, 1, 1, 0, 0, 0)
        ts = to_timestamp(dt)
        f = TimestampField()
        res = f.clean(ts)
        dt = dt.replace(tzinfo=timezone.utc)
        self.assertEqual(dt, res)

    def test_bounds(self):
        f = TimestampField(in_future=True)
        f.clean(0)
        f.clean(100000000)
        f.clean(2147483647)
        with self.assertRaises(ValidationError):
            f.clean(-random.randint(1, 2147483648))

        with self.assertRaises(ValidationError):
            f.clean(random.randint(2147483648, 4000000000))

    def test_in_future(self):
        f = TimestampField()
        future = now() + datetime.timedelta(hours=1)
        with self.assertRaises(ValidationError):
            f.clean(future)

    def test_initial(self):
        now_dt = now().replace(microsecond=0)
        f = TimestampField(initial=now_dt, required=False)
        res = f.clean(None)
        self.assertEqual(now_dt, res)
        f = TimestampField(initial=to_timestamp(now_dt), required=False)
        res = f.clean(None)
        self.assertEqual(now_dt, res)

    def test_required(self):
        f = TimestampField(required=False)
        self.assertEqual(None, f.clean(None))

        f = TimestampField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = TimestampField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class DateTimeFieldTest(TestCase):
    def test_now(self):
        now_dt = now().replace(microsecond=0)
        f = DateTimeField()
        res = f.clean(now_dt.strftime("%Y-%m-%dT%H:%M:%S"))
        self.assertEqual(now_dt, res)

    def test_initial(self):
        now_dt = now().replace(microsecond=0)
        f = DateTimeField(initial=now_dt, required=False)
        res = f.clean(None)
        self.assertEqual(now_dt, res)

    def test_required(self):
        f = DateTimeField(required=False)
        self.assertEqual(None, f.clean(None))

        f = DateTimeField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = DateTimeField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class DateFieldTest(TestCase):
    def test_today(self):
        today = date.today()
        f = DateField()
        res = f.clean(today.isoformat())
        self.assertEqual(today, res)

    def test_initial(self):
        today = date.today()
        f = DateField(initial=today, required=False)
        res = f.clean(None)
        self.assertEqual(today, res)

    def test_required(self):
        f = DateField(required=False)
        self.assertEqual(None, f.clean(None))

        f = DateField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = DateField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class MonthFieldTest(TestCase):
    def test_now(self):
        now_dt = now().replace(microsecond=0)
        f = MonthField()
        res = f.clean(now_dt.strftime("%Y-%m"))
        self.assertEqual(now_dt.date().replace(day=1), res)

    def test_initial(self):
        f = MonthField(initial='2017-01', required=False)
        res = f.clean(None)
        self.assertEqual(datetime.date(2017, 1, 1), res)

    def test_required(self):
        f = MonthField(required=False)
        self.assertEqual(None, f.clean(None))

        f = MonthField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = MonthField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class DateUnitFieldTest(TestCase):
    def test_day(self):
        f = DateUnitField()
        self.assertEqual('day', f.clean('day'))

    def test_hour(self):
        f = DateUnitField()
        self.assertEqual('hour', f.clean('hour'))

    def test_week(self):
        f = DateUnitField()
        self.assertEqual('week', f.clean('week'))

    def test_invalid(self):
        f = DateUnitField()
        with self.assertRaises(ValidationError):
            f.clean('something')

    def test_required(self):
        f = DateUnitField(required=False)
        self.assertEqual(None, f.clean(None))

        f = DateUnitField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        f = DateUnitField(required=False, initial='day')
        self.assertEqual('day', f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = DateUnitField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('day')


class ColorFieldTest(TestCase):
    def test_color(self):
        f = ColorField()
        self.assertEqual('afafaf', f.clean('afafaf'))

    def test_invalid(self):
        f = ColorField()
        with self.assertRaises(ValidationError):
            f.clean('test')

    def test_required(self):
        f = ColorField(required=False)
        self.assertEqual(None, f.clean(None))

        f = ColorField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        f = ColorField(required=False, initial='afafaf')
        self.assertEqual('afafaf', f.clean(None))


class TruncatedCharFieldTest(TestCase):
    def test_short_string(self):
        f = TruncatedCharField()
        self.assertEqual('afafaf', f.clean('afafaf'))

    def test_no_max(self):
        test_str = 't' * 10000
        f = TruncatedCharField(truncate_length=None)
        self.assertEqual(test_str, f.clean(test_str))

    def test_long_string(self):
        f = TruncatedCharField()
        self.assertEqual(f.clean('t' * 100500), 't' * 255)

    def test_required(self):
        f = TruncatedCharField(required=False)
        self.assertEqual(None, f.clean(None))

        f = TruncatedCharField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        f = TruncatedCharField(required=False, initial='afafaf')
        self.assertEqual('afafaf', f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = TruncatedCharField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class JsonFieldTest(TestCase):
    TEST_SCHEMA = {
        'type': 'object',
        'properties': {
            'test': {
                'type': 'integer',
                'minimum': 1
            }
        }
    }

    def test_json_correct(self):
        test_data = {'test': 1}
        f = JsonField(json_schema=self.TEST_SCHEMA)
        self.assertDictEqual(test_data, f.clean(json.dumps(test_data)))

    def test_json_incorrect(self):
        test_data = {'test': 0}
        f = JsonField(json_schema=self.TEST_SCHEMA)
        with self.assertRaises(ValidationError):
            f.clean(json.dumps(test_data))

    def test_required(self):
        f = JsonField(required=False)
        self.assertEqual(None, f.clean(None))

        f = JsonField()
        self.assertEqual({}, f.clean({}))
        self.assertEqual([], f.clean([]))
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = {'inital': True}
        f = JsonField(required=False, initial=test_data)
        self.assertEqual(test_data, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = JsonField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean({})


class ArrayFieldTest(TestCase):
    def test_json_valid(self):
        f = ArrayField()
        self.assertListEqual([1, 2, 3], f.clean('[1, 2, 3]'))

    def test_json_object(self):
        f = ArrayField()
        with self.assertRaises(ValidationError):
            f.clean('{}')

    def test_json_invalid(self):
        f = ArrayField()
        with self.assertRaises(ValidationError):
            f.clean('[1,2,3')

    def test_comma_separated(self):
        f = ArrayField()
        self.assertListEqual(['1', '2', '3'], f.clean('1,2,3'))

    def test_items_valid(self):
        test_data = [{'test': 1}, {'test': 2}]
        f = ArrayField(item_schema=JsonFieldTest.TEST_SCHEMA)
        self.assertListEqual(test_data, f.clean(json.dumps(test_data)))

    def test_items_invalid(self):
        test_data = [{'test': 1}, {'test': 0}]
        f = ArrayField(item_schema=JsonFieldTest.TEST_SCHEMA)
        with self.assertRaises(ValidationError):
            f.clean(json.dumps(test_data))

    def test_min_items_valid(self):
        test_data = [{'test': 1}]
        f = ArrayField(min_items=1)
        self.assertListEqual(test_data, f.clean(json.dumps(test_data)))

    def test_min_items_invalid(self):
        test_data = []
        f = ArrayField(min_items=1)
        with self.assertRaises(ValidationError):
            f.clean(json.dumps(test_data))

    def test_max_items_valid(self):
        test_data = [{'test': 1}, {'test': 1}]
        f = ArrayField(max_items=2)
        self.assertListEqual(test_data, f.clean(json.dumps(test_data)))

    def test_max_items_invalid(self):
        test_data = [{'test': 1}, {'test': 1}, {'test': 1}]
        f = ArrayField(max_items=2)
        with self.assertRaises(ValidationError):
            f.clean(json.dumps(test_data))

    def test_required(self):
        f = ArrayField(required=False)
        self.assertEqual(None, f.clean(None))

        f = ArrayField()
        self.assertEqual([], f.clean([]))
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = [1, 2, 3]
        f = ArrayField(required=False, initial=test_data)
        self.assertEqual(test_data, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = ArrayField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean([])


class IdArrayFieldTest(TestCase):
    def test_result(self):
        f = IdArrayField()
        self.assertListEqual([1, 2, 3, 4, 3, 1], f.clean('[1, 2, 3, 4, 3, 1]'))
        self.assertListEqual([1, 2, 3, 4, 3, 1], f.clean('["1", "2", "3", "4", "3", "1"]'))

        # Id can't be 0
        with self.assertRaises(ValidationError):
            f.clean('[0, 1, 3, 4]')

    def test_required(self):
        f = IdArrayField(required=False)
        self.assertIsNone(f.clean(None))

        f = IdArrayField()
        self.assertEqual([], f.clean([]))
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = [1, 2, 3]
        f = IdArrayField(required=False, initial=test_data)
        self.assertListEqual(test_data, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = IdArrayField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean([])


class IdSetFieldTest(TestCase):
    def test_result(self):
        f = IdSetField()
        self.assertSetEqual({1, 2, 3, 4}, f.clean('[1, 2, 3, 4, 3, 1]'))
        self.assertSetEqual({1, 2, 3, 4}, f.clean('["1", "2", "3", "4", "3", "1"]'))

        # Id can't be 0
        with self.assertRaises(ValidationError):
            f.clean('[0, 1, 3, 4]')

    def test_required(self):
        f = IdSetField(required=False)
        self.assertIsNone(f.clean(None))

        f = IdSetField()
        self.assertEqual(set(), f.clean([]))
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        test_data = {1, 2, 3}
        f = IdSetField(required=False, initial=test_data)
        self.assertEqual(test_data, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = IdSetField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean([])


class UrlFieldTest(TestCase):
    def test_url_valid(self):
        test_data = 'http://test.ru'
        f = UrlField()
        self.assertEqual(test_data, f.clean(test_data))

    def test_url_invalid(self):
        test_data = 'not_url'
        f = UrlField()
        with self.assertRaises(ValidationError):
            f.clean(test_data)

    def test_required(self):
        f = UrlField(required=False)
        self.assertEqual(None, f.clean(None))

        f = UrlField()
        with self.assertRaises(ValidationError):
            f.clean(None)

    def test_initial(self):
        f = UrlField(required=False, initial='http://test.ru')
        self.assertEqual('http://test.ru', f.clean(None))

    def test_regex(self):
        test_data = 'http://test.ru/test.jpg'
        f = UrlField(regex=r'^.*\.(jpg|jpeg|png|gif)*$', flags=re.I)
        self.assertEqual(test_data, f.clean(test_data))

        f = UrlField(regex=r'^.*\.txt$', flags=re.I)
        with self.assertRaises(ValidationError):
            f.clean(test_data)

    def test_underscore_domain_allowed(self):
        test_data = 'http://test_test.ru/test.jpg'
        f = UrlField()
        self.assertEqual(test_data, f.clean(test_data))

    def test_underscore_domain_rejected(self):
        test_data = 'http://test_test.ru/test.jpg'
        f = UrlField(with_underscore_domain=False)

        with self.assertRaises(ValidationError):
            self.assertEqual(test_data, f.clean(test_data))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = UrlField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class UUIDFieldTest(TestCase):
    def test_uuid_valid(self):
        test_data = str(uuid.uuid4())
        f = UUIDField()
        self.assertEqual(test_data, f.clean(test_data))

    def test_url_invalid(self):
        test_data = 'not_uuid'
        f = UUIDField()
        with self.assertRaises(ValidationError):
            f.clean(test_data)

    def test_required(self):
        f = UUIDField(required=False)
        self.assertEqual(None, f.clean(None))

    def test_initial(self):
        init = str(uuid.uuid4())
        f = UUIDField(required=False, initial=init)
        self.assertEqual(init, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = UUIDField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')


class FileFieldTest(TestCase):
    @staticmethod
    def _get_test_file(extension='txt'):
        f = BytesIO(b'test')
        f.name = 'test.' + extension
        f.size = len(f.getvalue())
        return f

    def test_file_valid(self):
        test_data = self._get_test_file()
        f = FileField()
        self.assertEqual(test_data, f.clean(test_data))

    def test_file_invalid(self):
        f = FileField()
        with self.assertRaises(ValidationError):
            f.clean('123')

    def test_file_extensions(self):
        with self.assertRaises(AssertionError):
            FileField(valid_extensions="test")

        test_file = self._get_test_file('pdf')

        f = FileField(valid_extensions=["pdf", "png"])
        self.assertEqual(test_file, f.clean(test_file))

        test_file.name = 'TEST.PDF'
        self.assertEqual(test_file, f.clean(test_file))

        f = FileField(valid_extensions=["png", "jpg"])
        with self.assertRaises(ValidationError):
            f.clean(f)

    def test_file_size(self):
        with self.assertRaises(AssertionError):
            FileField(max_size="test")

        test_file = self._get_test_file('pdf')

        f = FileField(max_size=2 * 1024 * 1024)
        self.assertEqual(test_file, f.clean(test_file))

        test_file.size = 1 * 1024 * 1024
        self.assertEqual(test_file, f.clean(test_file))

        test_file.size = 2 * 1024 * 1024 - 1
        self.assertEqual(test_file, f.clean(test_file))

        test_file.size = 2 * 1024 * 1024 + 1
        with self.assertRaises(ValidationError):
            f.clean(f)

    def test_required(self):
        f = FileField(required=False)
        self.assertEqual(None, f.clean(None))

    def test_initial(self):
        init = self._get_test_file()
        f = FileField(required=False, initial=init)
        self.assertEqual(init, f.clean(None))

    def test_empty_value_validators(self):
        # By default django skips run_validators methods, if value is in empty_values
        # It's not correct for REST, as empty value is not equal to None value now
        f = FileField(validators=[TestErrorValidator(0)])
        with self.assertRaises(ValidationError):
            f.clean('')



"""
This file contains a number of custom fields to validate data with django forms
"""

import json
from datetime import timezone, datetime, date
from typing import Any, Optional, Union

import os
import re
from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.forms import NullBooleanSelect, Textarea
from django.utils.timezone import make_aware, now

from .compatibility import to_timestamp, PatternType, string_types
from .exceptions import FileSizeError, FileTypeError
from .validators import URLValidatorWithUnderscoreDomain


class BaseField(forms.Field):
    """
    All library fields are inherited from this base
    Adds source
    """

    def __init__(self, *args, **kwargs):
        self.source = kwargs.pop('source', None)
        super(BaseField, self).__init__(*args, **kwargs)


class InitialFixMixin(BaseField):
    """
    Fixes problem with initial not returned by some form fields
    """

    def clean(self, value):  # type: (Any) -> Any
        value = super(InitialFixMixin, self).clean(value)
        return self.initial if value is None else value


class EmptyStringFixMixing(BaseField):
    """
    Fixes some fields error, returning empty string (not None) when value is not provided
    """
    # All empty values are considered by django as "No value" by default.
    # This causes a list of errors in REST model:
    # 1) JsonField and child classes don't allow to pass empty arrays and dicts if required=True
    # 2) RestCharField and child classes don't allow to pass empty string if required=True
    # 3) run_validators() method is not called if value is in empty_values, so min_length validation doesn't work
    #    for RestCharField and value=''
    empty_values = [None]

    def to_python(self, value):  # type: (Any) -> Optional[str]
        return None if value is None else super(EmptyStringFixMixing, self).to_python(value)

    def clean(self, value):  # type: (Any) -> Optional[str]
        """
        Returns initial value, when value is not provided
        :param value: value to clean
        :return: cleaned value
        """
        if value is None:
            value = self.to_python(value)
            self.validate(value)
            # BUG I don't call self.run_validators() here, as they don't expect None, but empty string
            return self.initial
        else:
            return super(EmptyStringFixMixing, self).clean(value)

    def validate(self, value):  # type: (Optional[str]) -> None
        """
        Fixes None value validation when required is True
        :param value: value to validate
        :return: None
        """
        if value is None and self.required:
            raise ValidationError(self.error_messages['required'], code='required')

        super(EmptyStringFixMixing, self).validate(value)


class RestCharField(EmptyStringFixMixing, forms.CharField):
    """
    Wraps django.forms.forms.CharField:
    + Changes default value - None, not empty string
    + Fixes initial value bug (CharField returns empty string, ignoring 'initial' parameter)
    """
    pass


class RegexField(RestCharField):
    """
    Wraps CharField to validate via regular expression
    """

    def __init__(self, *args, **kwargs):
        self.regex = kwargs.pop('regex', None)
        self.flags = kwargs.pop('flags', 0)

        assert self.regex is None or isinstance(self.regex, (string_types, PatternType)), \
            'regex must be string if given'
        assert isinstance(self.flags, int), 'flags must be integer'

        super(RegexField, self).__init__(*args, **kwargs)
        self._match = None

    def validate(self, value):  # type: (Optional[str]) -> None
        super(RegexField, self).validate(value)
        if value is not None and self.regex:
            self._match = re.match(self.regex, str(value), self.flags)
            if not self._match:
                raise ValidationError('Value don\'t match regexp "{0}"'.format(self.regex))

    @property
    def match(self):
        return self._match


class RestChoiceField(EmptyStringFixMixing, forms.ChoiceField):
    """
    Wraps django.forms.forms.ChoiceField:
    + Changes default value - None, not empty string
    + Fixes initial value bug (CharField returns empty string, ignoring 'initial' parameter)
    + Gives opportunity to set 'choices' as iterable of values, not iterable of tuples
    """

    def __init__(self, *args, **kwargs):
        choices = kwargs.pop('choices', None)
        if choices is not None:
            kwargs['choices'] = [ch if isinstance(ch, (list, tuple)) else (ch, ch) for ch in choices]

        EmptyStringFixMixing.__init__(self, *args, **kwargs)

        # This parameter was processed above, but was not removed
        kwargs.pop('source', None)

        forms.ChoiceField.__init__(self, *args, **kwargs)

    def validate(self, value):
        # Fixes bug with empty string passing choices
        if value is not None and self.choices and not self.valid_value(value):
            raise ValidationError(
                self.error_messages['invalid_choice'],
                code='invalid_choice',
                params={'value': value},
            )
        super(RestChoiceField, self).validate(value)


class RestIntegerField(InitialFixMixin, forms.IntegerField):
    """
    Wraps django.forms.forms.IntegerField, fixing 'initial' property ignore
    """
    pass


class RestFloatField(InitialFixMixin, forms.FloatField):
    """
    Wraps django.forms.forms.FloatField, fixing 'initial' property ignore
    """
    pass


class PositiveIntegerField(RestIntegerField):
    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param with_zero: If true, than integer can be equal to zero, else - not.
        :param kwargs: Named arguments
        """
        with_zero = kwargs.pop('with_zero', False)
        kwargs['min_value'] = kwargs.get('min_value', 0 if with_zero else 1)
        super(PositiveIntegerField, self).__init__(*args, **kwargs)


class IdField(PositiveIntegerField):
    """
    This field validates id (integer >= 1)
    """

    def __init__(self, *args, **kwargs):
        from django.conf import settings
        max_value = getattr(settings, "ID_FIELD_MAX_VALUE", None)
        kwargs['with_zero'] = kwargs.get('with_zero', False)
        if max_value is not None:
            kwargs['max_value'] = kwargs.get('max_value', max_value)

        super(IdField, self).__init__(*args, **kwargs)


class TimestampField(RestFloatField):
    """
    Form field, containing timestamp integer. Converts given value to datetime object
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes field.
        :param args: Positional arguments
        :param in_future: Boolean. If flag is false, validates, that given date is not greater than now
        :param kwargs: Named arguments
        """
        in_future = kwargs.pop('in_future', True)

        assert type(in_future) is bool, "in_future must be boolean"
        self._in_future = in_future

        # Converts initial to timestamp, if it is given as datetime
        if 'initial' in kwargs:
            assert isinstance(kwargs['initial'], (datetime, int, float)), \
                "initial must be int, float or datetime instance"
            dt = kwargs['initial']
            if isinstance(dt, datetime):
                kwargs['initial'] = to_timestamp(dt)

        super(TimestampField, self).__init__(*args, min_value=0, max_value=2147483647, **kwargs)

    def clean(self, value):  # type: (Any) -> Optional[datetime]
        value = super(TimestampField, self).clean(value)
        if value == 0:
            # Fix python 3.6 issue with Invalid argument value=0
            return datetime(1970, 1, 1, tzinfo=timezone.utc)
        elif value is not None:
            dt = datetime.utcfromtimestamp(value)
            return make_aware(dt, timezone.utc)
        else:
            return value

    def validate(self, value):  # type: (Optional[int]) -> None
        super(TimestampField, self).validate(value)
        if value is not None:
            if not self._in_future and value > to_timestamp(now()):
                raise ValidationError("You can't search time in future")


class DateTimeField(RestCharField):
    """
    Parses given string as datetime by given mask with datetime.strptime() method
    """
    base_type = datetime

    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param mask: Mask to parse datetime string with datetime.strptime() method
        :param kwargs: Named arguments
        """
        mask = kwargs.pop("mask", "%Y-%m-%dT%H:%M:%S")
        assert isinstance(mask, string_types), "'mask' parameter must be string"

        super(DateTimeField, self).__init__(*args, **kwargs)

        self.mask = mask

    def clean(self, value):  # type: (Any) -> datetime
        value = super(DateTimeField, self).clean(value)
        if value is not None and not isinstance(value, self.base_type):
            try:
                dt = datetime.strptime(value, self.mask)
            except (ValueError, TypeError):
                raise ValidationError("Invalid value format (%s)" % self.mask)

            return make_aware(dt, timezone.utc)
        else:
            return value


class DateField(DateTimeField):
    """
    Parses month with datetime.strptime() method. Default format is %Y-%m.
    Returns datetime.date with first day of month.
    """
    base_type = date

    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param mask: Mask to parse month string with datetime.strptime() method
        :param kwargs: Named arguments
        """
        mask = kwargs.pop("mask", "%Y-%m-%d")
        super(DateField, self).__init__(*args, mask=mask, **kwargs)

    def clean(self, value):  # type (Any) -> datetime.date
        value = super(DateField, self).clean(value)
        return value.date() if isinstance(value, datetime) else value


class MonthField(DateTimeField):
    """
    Parses month with datetime.strptime() method. Default format is %Y-%m.
    Returns datetime.date with first day of month.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param mask: Mask to parse month string with datetime.strptime() method
        :param kwargs: Named arguments
        """
        mask = kwargs.pop("mask", "%Y-%m")
        super(MonthField, self).__init__(*args, mask=mask, **kwargs)

    def clean(self, value):  # type (Any) -> date
        value = super(MonthField, self).clean(value)
        return value.date() if isinstance(value, datetime) else value


class TimezoneField(RestCharField):
    """
    A field containing one of standard pytz timezone names
    """

    def validate(self, value):  # type: (Optional[str]) -> None
        super(TimezoneField, self).validate(value)
        try:
            import pytz
        except ImportError as ex:
            raise ImportError('pytz is required  for TimezoneField') from ex

        if value and value not in pytz.all_timezones:
            raise ValidationError("Invalid timezone '{0}'".format(value))


class DateUnitField(RestChoiceField):
    """
    A field that validates one of date unit choices: hour, day or week
    """
    UNIT_CHOICES = ('hour', 'day', 'week')

    def __init__(self, *args, **kwargs):
        super(DateUnitField, self).__init__(*args, choices=self.UNIT_CHOICES, **kwargs)


class RestBooleanField(RestCharField):
    """
    Standard BooleanField is based on /django/forms/widgets.py CheckboxInput.value_from_datadict(value)
    It works improperly for REST model: required=True + value=False => ValidationError
    This filed fixes this issue, giving opportunity to send (required or not):
    + None as default value
    + 'false', '0', '' (ignoring case) as False
    + Everything else is parsed with standard bool function
    """
    widget = NullBooleanSelect

    def to_python(self, value):  # type: (Any) -> Optional[bool]
        """Returns a Python boolean object."""
        if value is None:
            return None
        elif isinstance(value, string_types) and value.lower() in ('false', '0', ''):
            value = False
        else:
            value = bool(value)
        return value


class LowerCaseEmailField(forms.EmailField):
    """
    Wraps django.forms.forms.EmailField:
    + Converts all emails to lowercase
    + Fixes initial value bug (CharField returns empty string, ignoring 'initial' parameter)
    + Changes default value - None, not empty string
    """

    def to_python(self, value):  # type: (Any) -> Optional[str]
        return value.lower() if isinstance(value, string_types) else value

    def clean(self, value):
        if value is not None:
            return forms.EmailField.clean(self, value)
        elif self.required:
            raise ValidationError(self.error_messages['required'], code='required')
        else:
            return self.initial


class ColorField(RestCharField):
    """
    This field validates color as six hex characters
    """

    def validate(self, value):  # type: (Optional[str]) -> None
        super(ColorField, self).validate(value)
        if value and not re.match('^[0-9a-f]{6}$', value):
            raise ValidationError("Color '{0}' is invalid".format(value))


class TruncatedCharField(RestCharField):
    """
    This field truncates string by given length
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param truncate_length: A positive integer, setting number of symbols to truncate input
        :param kwargs: Named arguments
        """
        truncate_length = kwargs.pop('truncate_length', 255)
        assert truncate_length is None or type(truncate_length) is int and truncate_length > 0, \
            "'truncate_length' parameter must be positive integer"

        self._truncate_length = truncate_length

        super(TruncatedCharField, self).__init__(*args, **kwargs)

    def to_python(self, value):  # type: (Any) -> Optional[str]
        if value and self._truncate_length is not None:
            value = value[:self._truncate_length]
        return super(TruncatedCharField, self).to_python(value)


class JsonField(RestCharField):
    """
    This field Json serialized string with jsonschema
    """
    widget = Textarea

    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param json_schema: jsonschema library validation schema
        :param kwargs: Named arguments
        """

        self._json_schema = kwargs.pop('json_schema', None)
        super(JsonField, self).__init__(*args, **kwargs)

    def to_python(self, value):  # type: (Any) -> Optional[Union[dict, list]]
        if value is None or isinstance(value, (list, dict)):
            return value
        elif isinstance(value, string_types):
            try:
                return json.loads(value)
            except Exception as e:
                raise ValidationError("Json was not parsed [{0}]".format(str(e)))
        else:
            raise ValidationError("Invalid JSON value [{0}]".format(str(value)))

    def validate(self, value):  # type: (Optional[Union[dict, list]]) -> None
        super(JsonField, self).validate(value)
        if self._json_schema and value is not None:
            try:
                import jsonschema
            except ImportError as ex:
                raise ImportError("jsonschema library is not required for this operation") from ex

            try:
                jsonschema.validate(value, self._json_schema)
            except jsonschema.exceptions.ValidationError as ex:
                raise ValidationError(ex.message)


class ArrayField(JsonField):
    """
    Field, that parses array from string. It can be passed in 2 forms:
    + strings, split by ',' symbol
    + json-encoded array
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes field
        :param args: Positional arguments
        :param item_schema: jsonschema dict to validate each item of the array
        :param min_items: Minimum number of items in the array
        :param max_items: Maximum number of items in the array
        :param kwargs: Named arguments
        """
        item_schema = kwargs.pop('item_schema', None)
        min_items = kwargs.pop('min_items', 0)
        max_items = kwargs.pop('max_items', None)
        json_schema = {
            "type": "array",
            "minItems": min_items
        }
        if max_items:
            json_schema['maxItems'] = max_items

        if item_schema:
            json_schema['items'] = item_schema

        kwargs['json_schema'] = json_schema
        super(ArrayField, self).__init__(*args, **kwargs)

    def to_python(self, value):  # type: (Any) -> Optional[list]
        if value is None or isinstance(value, list):
            return super(ArrayField, self).to_python(value)
        elif isinstance(value, dict):
            raise ValidationError('Value is expected to be JSON array, not object')
        elif isinstance(value, string_types):
            if value.startswith('[') or value.endswith(']'):
                return super(ArrayField, self).to_python(value)
            elif value.startswith('{') and value.endswith('}'):
                raise ValidationError('Value is expected to be JSON array, not object')
            else:
                # Comma separated list
                data = value.split(',')
                item_schema = self._json_schema.get('items') if self._json_schema else None
                if item_schema and item_schema.get('type') == 'integer':
                    try:
                        return [int(item) for item in data]
                    except (ValueError, TypeError):
                        raise ValidationError('Array of integers')
                else:
                    return data
        else:
            raise ValidationError("Invalid JSON value [{0}]".format(str(value)))


class IdArrayField(ArrayField):
    """
    Field, that parses array of ids from string. Each element is cleaned by IdField()
    """

    def clean(self, value):
        if isinstance(value, set):
            value = list(value)
        res = super(IdArrayField, self).clean(value)
        if isinstance(res, list):
            id_field = IdField()
            res = [id_field.clean(item) for item in res]
        return res


class IdSetField(IdArrayField):
    """
    Field, that parses set of ids from string. Duplicated ids are removed.
    Each element is cleaned by IdField()
    """

    def clean(self, value):
        res = super(IdSetField, self).clean(value)
        if isinstance(res, list):
            res = set(res)
        return res


class UrlField(RegexField):
    """
    Field validates url string
    """

    def __init__(self, *args, with_underscore_domain=True, **kwargs):  # type (bool) -> None
        super(UrlField, self).__init__(*args, **kwargs)
        self.with_underscore_domain = with_underscore_domain

    def to_python(self, value):  # type: (Any) -> Optional[str]
        if isinstance(value, string_types):
            value = value.strip()
        return super(UrlField, self).to_python(value)

    def validate(self, value):  # type: (Optional[str]) -> None
        super(UrlField, self).validate(value)
        if isinstance(value, string_types):
            validator_cls = URLValidatorWithUnderscoreDomain if self.with_underscore_domain else URLValidator
            v = validator_cls(schemes=['http', 'https'])
            v(value)


class HexField(RestCharField):
    """
    Field validates hexadecimal string
    """

    def validate(self, value):  # type: (Optional[str]) -> None
        super(HexField, self).validate(value)
        if isinstance(value, string_types) and not re.match(r'[a-z0-9]*', value):
            raise ValidationError("Field can contain only hexadecimal small characters (a-z, 0-9)")


class UUIDField(RegexField):
    """
    Field validates correct UUID string
    """

    def __init__(self, *args, **kwargs):
        kwargs['regex'] = r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
        super(UUIDField, self).__init__(*args, **kwargs)


class FileField(forms.FileField):
    """
    Wraps django.forms.forms.FileField, adding:
    + file size validation
    + file extension validation
    """

    def __init__(self, *args, **kwargs):
        max_size = kwargs.pop('max_size', None)
        valid_extensions = kwargs.pop('valid_extensions', None)

        self.valid_extensions = valid_extensions
        self.size = max_size

        if self.valid_extensions is not None:
            assert isinstance(self.valid_extensions, (tuple, set, list)), "valid_extensions must be tuple, list or set"
            assert len(self.valid_extensions) > 0, "valid_extensions can't be empty, if given"

        if self.size is not None:
            assert type(self.size) is int and self.size > 0, "size must be positive integer"

        super(FileField, self).__init__(*args, **kwargs)

    def validate(self, value):  # type (IO) -> None
        super(FileField, self).validate(value)
        if self.valid_extensions is not None and value:
            ext = os.path.splitext(value.name)[1][1:].lower()
            if ext not in self.valid_extensions:
                raise FileTypeError(ext, self.valid_extensions)

        if self.size is not None and value:
            if value.size > self.size:
                raise FileSizeError(self.size, value.size)

    def clean(self, data, initial=None):
        # FileField provides InitialFixMixin functionality with special parameter
        return super(FileField, self).clean(data, initial=initial if initial is not None else self.initial)



from unittest import TestCase

from django.forms import CharField
from django.db import models
from django_rest_form_fields import RestCharField, RestChoiceField
from django_rest_form_fields.forms import BaseForm, BaseModelForm


class ModelExample(models.Model):
    model_field = models.CharField(max_length=255)


class SourceBaseFormTest(TestCase):
    data = {
        'src': 'test'
    }

    def _test_form(self, base_cls, fields, result):
        FormExample = type('FormExample', (base_cls,), fields)

        f = FormExample(self.data)
        f.full_clean()
        self.assertDictEqual(result, f.cleaned_data)

    def test_source(self):
        self._test_form(BaseForm, {'dest': RestCharField(source='src')}, {'dest': 'test'})

    def test_no_source(self):
        self._test_form(BaseForm, {'src': RestCharField()}, {'src': 'test'})

    def test_native_field(self):
        self._test_form(BaseForm, {'src': CharField()}, {'src': 'test'})

    def test_required(self):
        self._test_form(BaseForm, {'dest': RestCharField(source='absent', required=False)}, {'dest': None})
        self._test_form(BaseForm, {'dest': RestCharField(source='src', required=False)},
                        {'dest': 'test'})

    def test_initial(self):
        self._test_form(BaseForm, {'dest': RestCharField(source='absent', required=False, initial='init')},
                        {'dest': 'init'})
        self._test_form(BaseForm, {'dest': RestCharField(source='src', required=False, initial='init')},
                        {'dest': 'test'})

    def test_choice_field(self):
        # The was a bug in RestChoiceField due to duplicate init
        self._test_form(BaseForm, {'dest': RestChoiceField(source='src', choices=('test',))}, {'dest': 'test'})


class SourceBaseModelFormTest(TestCase):
    data = {
        'src': 'test',
        'model_field': 'm'
    }

    def _test_form(self, base_cls, fields, result):
        fields['Meta'] = type('Meta', (), {'model': ModelExample, 'fields': ('model_field',)})
        FormExample = type('FormExample', (base_cls,), fields)

        f = FormExample(self.data)
        f.full_clean()
        self.assertDictEqual(result, f.cleaned_data)

    def test_source(self):
        self._test_form(BaseModelForm, {'dest': RestCharField(source='src')}, {'dest': 'test', 'model_field': 'm'})

    def test_no_source(self):
        self._test_form(BaseModelForm, {'src': RestCharField()}, {'src': 'test', 'model_field': 'm'})

    def test_native_field(self):
        self._test_form(BaseModelForm, {'src': CharField()}, {'src': 'test', 'model_field': 'm'})

    def test_required(self):
        self._test_form(BaseModelForm, {'dest': RestCharField(source='absent', required=False)},
                        {'dest': None, 'model_field': 'm'})
        self._test_form(BaseModelForm, {'dest': RestCharField(source='src', required=False)},
                        {'dest': 'test', 'model_field': 'm'})

    def test_initial(self):
        self._test_form(BaseModelForm, {'dest': RestCharField(source='absent', required=False, initial='init')},
                        {'dest': 'init', 'model_field': 'm'})
        self._test_form(BaseModelForm, {'dest': RestCharField(source='src', required=False, initial='init')},
                        {'dest': 'test', 'model_field': 'm'})


class RequiredBaseModelFormTest(TestCase):
    def test_base_form(self):
        class TestForm(BaseForm):
            test = RestCharField(max_length=255, required=False)

        f = TestForm({})
        f.full_clean()
        self.assertTrue(f.is_valid())
        self.assertDictEqual({'test': None}, f.cleaned_data)

    def test_model_form(self):
        class TestForm(BaseModelForm):
            class Meta:
                model = ModelExample
                fields = ('model_field',)

            model_field = RestCharField(max_length=255, required=False)

        f = TestForm({})
        f.full_clean()
        self.assertTrue(f.is_valid())
        self.assertDictEqual({'model_field': None}, f.cleaned_data)

    def test_model_form_initial(self):
        class TestForm(BaseModelForm):
            class Meta:
                model = ModelExample
                fields = ('model_field',)

            model_field = RestCharField(max_length=255, required=False, initial='init')

        f = TestForm({})
        f.full_clean()
        self.assertTrue(f.is_valid())
        self.assertDictEqual({'model_field': 'init'}, f.cleaned_data)



import re
from django.core.validators import URLValidator

try:
    # This attribute has been moved from URLValidator to DomainNameValidator in django 5.2
    from django.core.validators import DomainNameValidator

    ul = getattr(DomainNameValidator, "ul", getattr(URLValidator, "ul", None))
except ImportError:
    ul = URLValidator.ul



class URLValidatorWithUnderscoreDomain(URLValidator):
    hostname_re = r'[a-z' + ul + r'0-9](?:[a-z' + ul + r'0-9-_]{0,61}[a-z' + ul + r'0-9])?'
    host_re = '(' + hostname_re + URLValidator.domain_re + URLValidator.tld_re + '|localhost)'

    regex = re.compile(
        r'^(?:[a-z0-9.+-]*)://'  # scheme is validated separately
        r'(?:[^\s:@/]+(?::[^\s:@/]*)?@)?'  # user:pass authentication
        r'(?:' + URLValidator.ipv4_re + '|' + URLValidator.ipv6_re + '|' + host_re + ')'
        r'(?::\d{2,5})?'  # port
        r'(?:[/?#][^\s]*)?'  # resource path
        r'\Z', re.IGNORECASE
    )



"""
This file contains functions for different python and django version compatibility
"""
import datetime
import re
import sys
from datetime import timezone

from django.utils.timezone import make_aware

if hasattr(re, 'Pattern'):
    PatternType = re.Pattern
else:
    PatternType = re._pattern_type  # noqa

# six.string_types replacement in order to remove dependency
string_types = (str,) if sys.version_info[0] == 3 else (str, unicode)  # noqa F821


def to_timestamp(dt):  # type: (datetime.datetime) -> float
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        dt = make_aware(dt, timezone.utc)
    else:
        dt = dt.astimezone(timezone.utc)

    # dt.timestamp() does not work before python 3.3
    if hasattr(dt, 'timestamp'):
        return dt.timestamp()
    else:
        return (dt - datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)).total_seconds()
